/* 
 * phpssnap - Philips webcam capture program for FreeBSD/NetBSD.
 * Copyright (C) 2005 Jeff Forys <jeffware@marjum.com>
 *
 * Portions of this program were based on "phpsview".
 * Phpssnap abides by the same copyright, included below.
 */

const char *Version = "1.1 20051113";

const char *CopyrightVersion = "phpssnap %s\n\nCopyright (c) 2005 Jeff Forys.\nportions Copyright (c) 2003 Takafumi Mizuno\nportions Copyright (c) 1999-2003 Nemosoft Unv.\n\nRedistribution and use in source and binary forms, with or without\nmodification, is permitted under the terms specified in the source\ncode (and manual pages).\n";

#ifndef lint
static char rcsid[] = "$Id: phpssnap.c,v 1.6 2005/11/14 22:31:02 forys Exp $";
#endif

/*
 * Philips webcam Sequential Image Capture Program for NetBSD, version 0.0
 * phpsview - Sequential image capture program.
 * Copyright (C) 2003 Takafumi Mizuno <taka-qce@ls-a.jp>
 *
 * Portions of this program were modeled after or adapted from the
 * Philips webcam dirver for Linux by USB and Video4Linux interface part
 * (C) 1999-2003 Nemosoft Unv.; see
 * http://www.smcc.demon.nl/webcam/
 *
 * This program is free software; you can redistribute it and/or modify
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Reference: Image Processing Program by Linux,
 * author Iio Jun, ohmsha publishing
 *            ISBN:4-274-94623-1 ( Japanese Only! )
 */

#include <signal.h>
#include <sysexits.h>

/*
** Existing stuff from original "phpsview.c".
*/

#include "pwc.h"

#define	IFNUM 0

struct pwc_device	*pwcdev;
struct usb_device	usbdev;
unsigned char		*grabbuf;	/* native image format data */
unsigned char		*yuv;		/* YUV420P image data from above */

void			yuv420p2rgb(unsigned char *, unsigned char *);
extern int 		usb_pwc_init(void);
struct pwc_device *	usb_pwc_probe(struct usb_device *, unsigned int);

#define LOCAL_FPS       10
#define LOCAL_X         640
#define LOCAL_Y         480

#define LOCAL_SIZE_RGB	(LOCAL_X * LOCAL_Y * 3)
#define LOCAL_SIZE_YUV	(LOCAL_SIZE_RGB / 2)

#define MY_ENDP_SIZE	592	/* XXX My camera's endpoint max packet size */
#define	RAW_PKTSIZE	(LOCAL_SIZE_YUV+TOUCAM_HEADER_SIZE+TOUCAM_TRAILER_SIZE)
#define	STORE_BUFSIZE	(RAW_PKTSIZE*2+(MY_ENDP_SIZE*2))
				/* (MY_ENDP_SIZE*2) is margin */

/*
** New stuff for "phpssnap".
*/

char	*progname;
char	*opt_cmd[256];
char	*opt_usbdev = NULL;
int	opt_debug = 0;

int	Usage(const char *);
void	ErrorEx(const char *, const char *);
int	GetImage(unsigned char *);
void	PutImage(FILE *, unsigned char *);
void	Reaper(int);

int
main(int argc, char *argv[])
{
	unsigned char rgb[LOCAL_SIZE_RGB];	/* RGB image data */
	char dev[FILENAME_MAX];
	struct sigaction act;
	int cmdidx, i, ch, capival = 1000;

	/* set progname to path-stripped argv[0] */
	if ((progname = strrchr(argv[0], '/')) != NULL)
		progname++;
	else
		progname = argv[0];

	/* parse args */
	for (ch = 0; ch < sizeof(opt_cmd)/sizeof(*opt_cmd); ch++)
		opt_cmd[ch] = 0;

	cmdidx = 0;
	while ((ch = getopt(argc, argv, "c:di:u:v")) != -1) {
		switch (ch) {
		    case 'c':
			if (cmdidx < sizeof(opt_cmd)/sizeof(*opt_cmd))
				opt_cmd[cmdidx++] = optarg;
			else
				return Usage("too many commands");
			break;
		    case 'd':
			opt_debug++;
			break;
		    case 'i':
			capival = atoi(optarg);
			break;
		    case 'u':
			opt_usbdev = optarg;
			break;
		    case 'v':
			printf(CopyrightVersion, Version);
			exit(EX_OK);
			break;
		    default:
			return Usage("unknown flag");
		}
	}
	argc -= optind;
	argv += optind;

	if (argc != 0)
		return Usage(argv[0]);


	if (opt_debug)
		fprintf(stderr, "%s: args parsed, debug enabled (%d)\n",
			progname, opt_debug);

	if ((grabbuf = malloc(RAW_PKTSIZE)) == NULL ||
	    (yuv = malloc(LOCAL_SIZE_YUV)) == NULL)
		ErrorEx("malloc", "out of memory");

	usbdev.fd = -1;
	usbdev.efd = -1;
	pwcdev = NULL;

	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
	act.sa_handler = SIG_IGN;
	sigaction(SIGHUP, &act, (struct sigaction *)NULL);

	sigaddset(&act.sa_mask, SIGINT);
	sigaddset(&act.sa_mask, SIGQUIT);
	sigaddset(&act.sa_mask, SIGTERM);
	act.sa_handler = Reaper;
	sigaction(SIGINT, &act, (struct sigaction *)NULL);
	sigaction(SIGQUIT, &act, (struct sigaction *)NULL);
	sigaction(SIGTERM, &act, (struct sigaction *)NULL);

	usb_pwc_init();			/* initialize module */

	if (opt_usbdev != NULL) {
		strcpy(dev, opt_usbdev);
		if ((usbdev.fd = open(dev, O_RDWR)) < 0)
			ErrorEx(dev, strerror(errno));
		if ((pwcdev = usb_pwc_probe(&usbdev, IFNUM)) == NULL )
			ErrorEx(dev, "failed to identify camera");
	} else {
		for (i = 0; i < 128; i++) {
#ifdef __FreeBSD__
			sprintf(dev, "/dev/ugen%d", i);
#else
			sprintf(dev, "/dev/ugen%d.00", i);
#endif
			if ((usbdev.fd = open(dev, O_RDWR)) < 0)
				continue;

			if ((pwcdev = usb_pwc_probe(&usbdev, IFNUM)) != NULL)
				break;

			close(usbdev.fd);
			usbdev.fd = -1;
		}
	}

	if (usbdev.fd < 0)
		ErrorEx("/dev/ugen*", "failed to locate camera");

	if (pwc_video_open(pwcdev) < 0 ) {
		usb_pwc_disconnect(&usbdev, (void *)pwcdev);
		ErrorEx(dev, "failed to open camera");
	}

	if (pwc_try_video_mode(pwcdev, LOCAL_X, LOCAL_Y, LOCAL_FPS, 0, 0) != 0){
		usb_pwc_disconnect(&usbdev, (void *)pwcdev);
		ErrorEx(dev, "failed to initiate video mode");
	}

	/* open endpoint  */
#ifdef __FreeBSD__
	sprintf(usbdev.epdevname, "%s.%d", dev, pwcdev->vendpoint);
#else
	sprintf(usbdev.epdevname, "%s.%02d", strtok(dev,"."), pwcdev->vendpoint);
#endif
	if ((usbdev.efd = open(usbdev.epdevname, O_RDONLY)) < 0) {
		perror(usbdev.epdevname);
		pwc_video_close(pwcdev);
		usb_pwc_disconnect(&usbdev, (void *)pwcdev);
		ErrorEx(dev, "failed to open endpoint");
	} 

	/* image clear */
	memset((void *)grabbuf, 0, RAW_PKTSIZE);

	for (;;) {
		struct timeval tv_s, tv_f, tv_e;
		FILE *fp;

		gettimeofday(&tv_s, NULL);

		if (GetImage(rgb) != 0) {
			fp = NULL;

			for (i = 0; i < cmdidx; i++) {
				if (opt_debug > 1)
					fprintf(stderr, "%s: launch %d: %s\n",
						progname, i, opt_cmd[i]);
				fp = popen(opt_cmd[i], "w");
				if (fp != NULL) {
					PutImage(fp, rgb);
					pclose(fp);
				}
			}

			if (i == 0)
				PutImage(stdout, rgb);
		}
		/* how long to sleep for next iteration? */
		gettimeofday(&tv_f, NULL);
		timersub(&tv_f, &tv_s, &tv_e);
		tv_s.tv_sec = capival / 1000;
		tv_s.tv_usec = (capival % 1000) * 1000;
		if (timercmp(&tv_s, &tv_e, > )) {
			timersub(&tv_s, &tv_e, &tv_f);
			if (opt_debug > 1)
				fprintf(stderr, "%s: sleep %lu sec %lu usec\n",
					progname, tv_f.tv_sec, tv_f.tv_usec);
			if (tv_f.tv_sec > 0)
				sleep(tv_f.tv_sec);
			if (tv_f.tv_usec > 0)
				usleep(tv_f.tv_usec);
		}
	}
	/*NOTREACHED*/

	return 0;
}

int
Usage(const char *msg)
{
	fprintf(stderr, "%s: invalid argument: %s specified\n", progname, msg);
	fprintf(stderr, "%s [-dv] [-i <msec>] [-u <usbdev>] { [-c <cmd>] }\n",
		progname);
	return EX_USAGE;
}

void
ErrorEx(const char *what, const char *why)
{
	Reaper(0);
	fprintf(stderr, "%s: %s: %s\n", progname, what, why);
	exit(EX_OSERR);
}

void
Reaper(int sig)
{
	fprintf(stderr, "%s: closing %s... ", progname,
		pwcdev == NULL? "down": pwcdev->vdev_name);
	if (usbdev.efd >= 0)
		close(usbdev.efd);
	if (pwcdev != NULL)
		pwc_video_close(pwcdev);
	if (usbdev.fd >= 0) {
		usb_pwc_disconnect(&usbdev, (void *)pwcdev);
		close(usbdev.fd);
	}
	fprintf(stderr, "done.\n");
	exit(0);
}

void
PutImage(FILE *fp, unsigned char *rgb)
{
	int i;

	fprintf(fp, "P6\n");
	fprintf(fp, "%d %d 255 ", LOCAL_X, LOCAL_Y);

	for ( i = 0; i < (LOCAL_SIZE_RGB); i+=3 )
		fprintf(fp, "%c%c%c", rgb[i], rgb[i+1], rgb[i+2]);
}

/*
 * GetImage() - Read raw-data from camera and convert to RGB24.
 */
int
GetImage(unsigned char *rgb)
{
	unsigned char buf[STORE_BUFSIZE];
	const char *usbstr = "read(usbdev)";
	unsigned char *cp_e, *cp;
	ssize_t n, len;
	int ffd = 0;			/* opt_debug */
	u_int count = 0;		/* opt_debug */

	if (opt_debug > 3)
		ffd = open("/tmp/phpssnap.raw",O_WRONLY|O_CREAT|O_APPEND,0666);

	/*
	 * Read until we find a header.
	 */
	for (cp = cp_e = buf;
			(n=read(usbdev.efd, cp_e, sizeof buf - (cp_e-cp))) > 0;
			cp = buf, cp_e = buf + TOUCAM_HEADER_SIZE) {
		if (opt_debug > 3 && ffd >= 0) {
			write(ffd, cp_e, n);
			count += n;
		}
		if (opt_debug > 2)
			fprintf(stderr, "%s: hdr read %d bytes\n", progname, n);
		for (cp_e += n; cp < cp_e; cp++) {
			if ((cp+TOUCAM_HEADER_SIZE+1) == cp_e ||/* end of buf */
			    (*cp == 0x00 && *(cp+1) == 0x01 &&	/* compressed */
			     *(cp+6) == 0x04 &&
			     (*(cp+2) == *(cp+3) || *(cp+2) == 0x00)) ||
			    (*cp == 0x20 && *(cp+1) <= 0x02 &&      /* uncomp */
			     (*(cp+2) == 0x00 || *(cp+2) == 0xff) &&
			     (*(cp+4) <= 0x04) &&
			     (*(cp+6) == 0x04))) {
				if (opt_debug > 3 && ffd >= 0) {
					char buf[32];
					count -= (cp_e - cp);
					sprintf(buf, "\n=== %u ===\n", count);
					write(ffd, buf, strlen(buf));
					close(ffd);
				}
				break;
			}
		}

		if ((cp + TOUCAM_HEADER_SIZE + 1) != cp_e) {
			if (opt_debug > 1)
				fprintf(stderr,
					"%s: hdr %x %x %x %x %x %x %x %x\n",
					progname, *cp, *(cp+1) & 0xff,
					*(cp+2) & 0xff, *(cp+3) & 0xff,
					*(cp+4) & 0xff, *(cp+5) & 0xff,
					*(cp+6) & 0xff, *(cp+7) & 0xff);
			cp += TOUCAM_HEADER_SIZE;
			break;
		}

		memcpy((void *)buf, (void *)cp, TOUCAM_HEADER_SIZE);
	}

	if (n > 0) {
		if (opt_debug)
			fprintf(stderr, "%s: found header\n", progname);

		if ((n = cp_e - cp) > 0)
			memcpy(grabbuf, cp, n);

		for (cp = grabbuf + n, len = LOCAL_SIZE_YUV - n;
				len > 0; cp += n, len -= n) {
			if ((n = read(usbdev.efd, cp, len)) <= 0)
				break;
			if (opt_debug > 3)
				fprintf(stderr, "%s: read %d bytes\n",
					progname, n);
		}
		if (n > 0) {
			pwc_decompress(pwcdev);
			yuv420p2rgb(yuv, rgb);
		}
	}
	if (n < 0)
		ErrorEx(usbstr, strerror(errno));
	else if (n == 0)
		ErrorEx(usbstr, "end of file");

	return 1;
}

void yuv420p2rgb( unsigned char *yuvbuf, unsigned char *rgbbuf )
{
	int i, j;
	int halfoff;
	unsigned char *yp, *up, *vp;
	float y, u, v;
	float r, g, b;

	yp = yuvbuf;
	up = yuvbuf + LOCAL_X * LOCAL_Y;
	vp = up + (LOCAL_X/2) * (LOCAL_Y/2);

	for ( i = 0; i < LOCAL_Y; i++ ) {
		for ( j = 0; j < LOCAL_X; j++ ) {
			y = (float)*(yp + ( i * LOCAL_X + j ));
			halfoff = (((i%2==1)?i-1:i)/2)*(LOCAL_X/2) + ((j%2==1)?j-1:j)/2;
			u = (float)*(up + halfoff);
			v = (float)*(vp + halfoff);

			/* convert to RGB from YUV */
			u = u - 128.0;
			v = v - 128.0;

			r = 1.4*u + y;
			g = 1.02 * y - 0.750 * u - 0.336 * v;
			b = 1.77*v + y;

			if ( r > 255 ) r = 255;
			if ( r < 0 ) r = 0;
			if ( g > 255 ) g = 255;
			if ( g < 0 ) g = 0;
			if ( b > 255 ) b = 255;
			if ( b < 0 ) b = 0;

			/* print data */
			*rgbbuf++ = (unsigned char)b;
			*rgbbuf++ = (unsigned char)g;
			*rgbbuf++ = (unsigned char)r;
		}
	}

	return;
}
