/*
** Copyright (c) 2002, 2003, 2005
**      Jeff Forys (jeffware@marjum.com).  All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that: (1) Redistributions of
** source code must retain the above copyright notice, this list of
** conditions and the following disclaimer, (2) Redistributions in
** binary form must reproduce the above copyright notice, this list
** of conditions and the following disclaimer in the documentation
** and/or other materials provided with the distribution, (3) All
** advertising materials mentioning features or use of this software
** must display the following acknowledgment: ``This product includes
** software developed by Jeff Forys (jeffware@marjum.com).'', (4)
** The name of the author may not be used to endorse or promote products
** derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
** WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
** MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/

/*
 * Include file for applications that need to parse the config file.
 *
 * $Id: parser.h,v 1.11 2005/09/22 04:53:17 forys Exp $
 */

#include <sys/param.h>

/*
 * Configuration file constants, parse structure and external functions.
 */
#define	CONFLINELEN	1024		/* maximum length of config file line */
#define	CONFMSGLEN	256		/* maximum length of config error msg */
#define	CONFIMGLEN	256		/* maximum length of image file name */
#define	CONFCAMMAX	8		/* maximum number of unique cameras */
#define	CONFIMGMAX	32		/* maximum number of images/camera */

struct confdata_s {
	/* confread: vars filled in from config file */
	char *camdir;			/* (req) CAMDIR */
	char *curimgdir;		/* (req) CURIMGDIR */
	char *defimgdir;		/* (opt) DEFIMGDIR */
	char *curcondir;		/* (opt) CURCONDIR */
	char *cliapp;			/* (opt) CLIAPP */
	char *pidfil;			/* (req) PIDFIL */
	char *logfil;			/* (opt) LOGFIL */
	char *camusers;			/* (req) CAMUSERS */
	int usertout;			/* (opt) USERTOUT */
	int capival;			/* (opt) CAPIVAL */
	int capdelay;			/* (opt) CAPDELAY */
	int sampling;			/* (opt) SAMPLING */
	struct confcam_s {		/* (opt) CAMERA / CAPTURE / autoconf */
		char *name;					/* CAMERA */
		struct confcam_s *ovride;			/* CAMERA */
		char *imgfil[CONFIMGMAX], **imgfil_e;		/* CAMERA */
		char *capcmd;					/* CAPTURE */

		/* autoconf: additional per-camera vars used by 'cammgr' */
		void (*capfun)(const struct confcam_s *); /* capture func */
		char **defimg;				/* def image data */
		size_t *defsiz;				/* def image sizes */
		unsigned short ovcnt;			/* overriding cam cnt */
		unsigned short ovdone;			/* already overriden */
	} camera[CONFCAMMAX], *camera_e;
};

void			confinit(void);
struct confdata_s *	confread(const char *, const char **);
